// Copyright(c) 2005 Analog Devices, Inc. All Rights Reserved.
// This software is proprietary and confidential to Analog Devices, Inc. and its licensors.

// File    : $Id: //depot/development/visualaudio/common/layoutsupport/2.5.0/Include/AudioProcessing.h#3 $ 
// Part of : VisualAudio LayoutSupport Library
// Updated : $Date: 2006/10/12 $ by $Author: Fernando $

/**
 * AudioProcessing.h
 * Definitions for AMF module structures and AMF_Grouping.
 *
 *  Description:    Base instance, class and grouping structs for
 *                  post-processing audio modules
 *  Copyright:      Analog Devices (c) 2003
 *  Owner:          Analog Devices
 *
 *  Revision:
 *
 *  D. Jaffe        1 May 2003      Initial Revision
 *  D. Nixon        3 May 2003      Changed to use framework support functions, and be lintable.
 *                                  Added lots of comments.
 *  D. Jaffe        7 May 2003      Renamed to AudioPostProcessing.h.  Changed some fields, etc.
 *  D. Jaffe        27 May 2003     Added asm support, more segments.
 *  S. Costello     23 July 2003    Changed all references of AMF_ModuleInstance* to AMF_Module* 
 *  D. Jaffe        20 Apr 2004     Renamed to AudioProcessing.h.  
 *  D. Jaffe        17 Aug 2005     Factored to also support stand-alone module library
 */

#ifndef AUDIO_POST_PROCESSING_H
#define AUDIO_POST_PROCESSING_H

#ifdef __ECC__  // C compilation

// floating to fixed point compile-time macros. Note that for full precision, you must
//    specify the float with the "L" suffix, e.g. 0.736L
#define AMF_Clip32(x) ((x) > 2147483647.0L ? 2147483647.0L : ((x) < -2147483648.0L ? -2147483648.0L : (x)))
#define AMF_Fract32(x) ((int)AMF_Clip32(((x) * 2147483648) + 0.5L))

#define AMF_Clip16(x) ((x) > 32767.0L ? 32767.0L : ((x) < -32768.0L ? -32768.0L : (x)))
#define AMF_Fract16(x) ((short)AMF_Clip16(((x) * 32768) + 0.5L))

#ifdef __ADSPBLACKFIN__
#include <fract.h>
#define AMF_SIGNALS_ARE_FLOATING_POINT 0
typedef fract32 AMF_Signal;
#define AMF_Fract32ToSignal(x) (x)
#define AMF_FloatToSignal(x) AMF_Fract32(x)
#define AMF_FloatToShort(x) AMF_Fract16(x)
#ifndef pm
#define pm // This SHARC-ism is ignored for Blackfin
#endif
#else // SHARC
#define AMF_SIGNALS_ARE_FLOATING_POINT 1
typedef int fract32;
typedef float AMF_Signal;
#define AMF_Fract32ToSignal(x) __builtin_conv_RtoF((x))
#define AMF_FloatToSignal(x) ((x))
#define AMF_FloatToShort(x) ((x))
#endif

// Frequency-domain hop decision logic, argument is an AMF_GroupingStruct pointer
#define AMF_FD_ISHOP(x) ((!(x)) || ((x) && ((x)->freqDomainInfo->ticksToNextHop==0))) 

/** Memory segment to place render functions. */
#define SEG_MOD_FAST_CODE           section("mod_fastcode")

/** Memory segment for helper and initialization functions */
#define SEG_MOD_SLOW_CODE           section("mod_slowcode")

/** General data segment. */
#define SEG_MOD_FAST_SDATA          section("mod_fastsdata")

/** PM data segment (and type). */
#define SEG_MOD_FAST_SDATA_PM       pm section("mod_fastsdatapm")

/** Data segment that can be slow */
#define SEG_MOD_SLOW_SDATA          section("mod_slowsdata")

/** Data segment for constant data */
#define SEG_MOD_FAST_CONST          section("mod_fastconst")

/** Data segment (and type) for constant pm data */
#define SEG_MOD_FAST_CONST_PM       pm section("mod_fastconstpm")

/** Data segment for initialization data (can be slow access). */
#define SEG_MOD_SLOW_CONST          section("mod_slowconst")

/****** The following is used in VA 2.0.x and earlier only.  **/
    #define SEG_MOD_SRAM_SDATA          section("mod_sramsdata")
    #define SEG_PRESETS                 section("presets")
/****** End of VA 2.0.x and earlier only block. **/

struct AMF_ModuleStruct;            /* forward decl */
struct AMF_GroupingStruct;          /* forward decl */

/**
 * Module render function.
 *
 * This typedef defines the module render function prototype. The module
 * render functions are called with instance data and a buffer pointer
 * that includes input pins, output pins and scratch buffers.
 *
 * @param instance      Instance information for the render function.
 * @param buffers       Audio buffer pointers.
 * @param tickSize      Number of time samples to compute.
 */
typedef void (*AMF_RenderFunction)(
    struct AMF_ModuleStruct *restrict instance,
    AMF_Signal *restrict *buffers,
    int tickSize);

/** Frequency domain signal types are available in VisualAudio 2.5 and later.
 ** Be sure to set module XML's "layoutSupportLibraryVersion" to 2 (or latest
 ** revision) if you use these pin types.
 **/
typedef enum {
    AMFPinType_MONO = 0,
    AMFPinType_STEREO = 1,
    AMFPinType_CONTROL = 2,
    AMFPinType_SPECTRUM_REAL = 3,
    AMFPinType_SPECTRUM_COMPLEX = 4,
    AMFPinType_SPECTRUM_HALF_REAL = 5,
    AMFPinType_SPECTRUM_HALF_COMPLEX = 6,
    AMFPinType_PIN_TYPE_COUNT
} AMFPinType;

/*
 * Input pins' and output pins' type designation is done with
 * bit vectors. This macros below help document this.
 */
/**
 * Stereo bit for a given (zero-based) pin.
 */
#define AMF_StereoPin(whichPin) (AMFPinType_STEREO<<(whichPin*4))

/**
 * Control bit for a given (zero-based) pin.
 */
#define AMF_ControlPin(whichPin) (AMFPinType_CONTROL<<(whichPin*4))

/**
 * Mono bit for a given (zero-based) pin.
 */
#define AMF_MonoPin(whichPin) (0)

/**
 * Spectrum real bit for a given (zero-based) pin.
 */
#define AMF_SpectrumRealPin(whichPin) (AMFPinType_SPECTRUM_REAL<<(whichPin*4))

/**
 * Spectrum complex bit for a given (zero-based) pin.
 */
#define AMF_SpectrumComplexPin(whichPin) (AMFPinType_SPECTRUM_COMPLEX<<(whichPin*4))

/**
 * Spectrum half real bit for a given (zero-based) pin.
 */
#define AMF_SpectrumHalfRealPin(whichPin) (AMFPinType_SPECTRUM_HALF_REAL<<(whichPin*4))

/**
 * Spectrum half complex bit for a given (zero-based) pin.
 */
#define AMF_SpectrumHalfComplexPin(whichPin) (AMFPinType_SPECTRUM_HALF_COMPLEX<<(whichPin*4))


#define AMFModuleClassFlag_INDIRECT_INPUT_PIN_TYPE (1<<1)
#define AMFModuleClassFlag_INDIRECT_OUTPUT_PIN_TYPE (1<<2)

/**
 * Instance data for a post-processing module.
 *
 * This is the base class for all audio post-processing modules.
 */
typedef struct AMF_ModuleStruct
{
    /** Reference to static class descriptor.
     *  May actually be a AMF_ModuleVariableClass pointer
     *
     *  Need not be said when not using module self-description.
     */
    const struct _AMF_ModuleClass *classInfo;

    /** Module's table of indexes. This is used by the grouping object to resolve buffer
     *  pointers before the module's Render function is called.
     *  May also be used to store direct pointers to module buffers.
     *
     *  Need not be said when not using module self-description.
     */
    int *const bufferIndecies;

    /** The size of the array pointed to by bufferIndecies.
     *
     *  Need not be said when not using module self-description.
     */
    int bufferIndexCount;

    /** Every module (except the top-level grouping) is part of a grouping.
     *  Groupings may be hierarchical.
     *
     *  Need not be said when not using module self-description.
     */
    struct AMF_GroupingStruct * const grouping;

    /** Current function
     *
     *  Unused when not using run-time support and when module doesn't change
     *  its own render function (if it does, this is documented in module documentation.)
     */
    AMF_RenderFunction currentRenderFunc;
}
AMF_Module;

/* This macro is useful for stand-alone usage of modules, where run-time info is not needed. */
#define AMF_Module_NULL_INIT {NULL,NULL,0,NULL,NULL} 


/**
 * Instance data for variable I/O post-processing modules.
 *
 * This is the base class for audio post-processing modules that have variable
 * number of inputs or outputs.
 */
typedef struct
{
    /** Base class. */
    AMF_Module b;

    /** Actual number of inputs for this instance. */
    int inputPinCount;

    /** 
     *  The class description version is ignored.
     */
    unsigned int inputPinBitVectorOrPointer;

    /** Actual number of outputs for this instance. */
    int outputPinCount;
    
    /** 
     *  The class description version is ignored.
     */
    unsigned int outputPinBitVectorOrPointer;
}
AMF_ModuleVariable;


/*********************************************************************************
 * The following structures are needed by the VisualAudio layout support run-time.
 * They may be ignored when using the module library in a stand-alone fashion.
 */

/**
 * Const environment description for a post-processing module
 * that has a variable number of inputs and outputs.
 *
 * All members are constant.
 */
typedef struct
{
    /** Any flags describing this class of modules.
     *  See below for interpretation of this field.
     */
    const unsigned int flags;

    /** Render function used by the instances to render. */
    const AMF_RenderFunction renderFunction;

    /** If NULL, bypass is "straight accross" (see documentation),
     * otherwise this function performs the bypass.
     */
    const AMF_RenderFunction bypassFunction;
}
AMF_ModuleVariableClass;

/**
 * Const environment description for a post-processing module that has a fixed number of inputs and outputs.
 *
 * All members are constant. Note that first three fields are identical to
 * AMF_ModuleVariableClass, allowing safe casts. 
 */
typedef struct _AMF_ModuleClass
{
    /** Any flags describing this class of modules.
     *  See below for interpretation of this field.
     */
    const unsigned int flags;

    /** Render function used by the instances to render.
     */
    const AMF_RenderFunction renderFunction;

    /** If NULL, bypass is "straight accross" (see documentation),
     * otherwise this function performs the bypass.
     */
    const AMF_RenderFunction bypassFunction;

    /** The number of input pins (for fixed number of pins).
     *  Ignored for variable number of pins.
     */
    const int inputPinCount;

    /**
     * Ignored for variable number of pins.
     */
    const unsigned inputPinBitVectorOrPointer;

    /** The number of output pins (if fixed number of pins).
     *  Ignored for variable number of pins.
     */
    const int outputPinCount;

    /**
     *  Ignored for variable number of pins.
     */
    const unsigned outputPinBitVectorOrPointer;
}
AMF_ModuleClass;

/* Interpretation of AMF_ModuleClass and AMF_ModuleClassVariable flags field. */
/**
 * If set, each instance of the module supports a settable (at init time) number of pins.
 */
#define AMFModuleClassFlag_VARIABLE_PIN_COUNT    1

/**
 * This field specifies the version of the LayoutSupport run-time support that the
 * module assumes. The version info is checked by the run-time.
 * See module documentation for information on correct value for this field. 
 */
#define AMFModuleClassFlag_LAYOUT_SUPPORT_VERSION_MASK 0x00FF0000


/* Structure pointed to by Grouping */
/** Frequency domain signal types are available in VisualAudio 2.5 and later.
 ** Be sure to set module XML's "layoutSupportLibraryVersion" to 2 (or 
 ** latest revision) if you use these pin types.
 **/
typedef struct AMF_FreqDomainInfoStruct
{
    /* FFT size for this grouping (in terms of number of complex or real points) */
    int fftSize;

    /* Ticks per frequency domain hop */
    int ticksPerHop;

    /* Counter that counts down each tick */
    int ticksToNextHop;
    
    /* Reserved for future use */
    int reserved;
}
AMF_FreqDomainInfo;


/**
 * Post-processing module group description.
 *
 * This is the instance class for the grouping module.
 */
typedef struct AMF_GroupingStruct
{
    /** Base class. Since each group may have a variable number of pins, it inherits from
     *  AMF_ModuleVariable.
     */
    AMF_ModuleVariable b;

    /** Tick size for all modules of this grouping. */
    const int tickSize;

    /** If this module does a down-sampling or up-sampling by a power of 2,
     *  this is the ratio relative to parent. If no down or up-sampling is
     * done, this value is 1.0 */
    const float samplingRateRatio;

    /** The number of modules in the grouping. */
    const int numModules;

    /** Array of module references. */
    AMF_Module * const * const modules;

    /** Freq domain info */
    AMF_FreqDomainInfo * freqDomainInfo;

    /** Array describing initial status of each module, packed in 2 bit units,
     *  where the least significant 2 bits represents the initial status of the
     *  first module, the second least significant 2 bits represents the initial
     *  status of the second module, etc.
     *  If omitted, all modules default to active status
     *  The size of the array pointed to is ceil(numModules/16).
     */
    const unsigned int * const initialStatusBitVectors;

    /** This is set at initialization time and is used for memory allocation */
    int maxBufferCount;

    /**
     * Pointer to the layout.  
     */
    struct _AMF_Layout * const layout;
}
AMF_Grouping;

#else   // Assembly language defines


#ifdef __ADSPBLACKFIN__
#else // SHARC
//Firing logic macro for FD assembly fcns
#define FiringLogic i4=r4;\
r0=dm(grouping,i4);\
r0 = pass r0;\
if EQ jump _execute;\
i4=r0;\
i12=dm(freqDomainInfo,i4);\
i4=i12;\
r2=dm(ticksToNextHop,i4);\
r2 = pass r2;\
if EQ jump _execute;\
jump _dontExecute;\
_execute:
#endif


#define AMF_Module_SIZE         5

#define AMF_Module_classInfo_ptr    0
#define AMF_Module_buffers_ptr      1
#define AMF_Module_buffer_count     2
#define AMF_Module_grouping_ptr     3   
#define AMF_Module_render_function  4

#define AMF_ModuleVariable_SIZE     (AMF_Module_SIZE+4)

#define AMF_ModuleVariable_inputPinCount              (AMF_Module_SIZE+0)
#define AMF_ModuleVariable_inputPinIsStereoBitVector  (AMF_Module_SIZE+1)    
#define AMF_ModuleVariable_outputPinCount             (AMF_Module_SIZE+2)
#define AMF_ModuleVariable_outputPinIsStereoBitVector (AMF_Module_SIZE+3)   



/** Memory segment to place render functions. */
#define SEG_MOD_FAST_CODE           mod_fastcode

/** Memory segment for helper and initialization functions */
#define SEG_MOD_SLOW_CODE           mod_slowcode

/** General data segment. */
#define SEG_MOD_FAST_SDATA          mod_fastsdata

/** PM data segment. */
#define SEG_MOD_FAST_SDATA_PM       mod_fastsdatapm

/** Data segment that can be slow access */
#define SEG_MOD_SLOW_SDATA          mod_slowsdata

/** Data segment for constant data */
#define SEG_MOD_FAST_CONST          mod_fastconst

/** Data segment for constant pm data */
#define SEG_MOD_FAST_CONST_PM       mod_fastconstpm

/** Data segment for initialization data (can be slow access). */
#define SEG_MOD_SLOW_CONST          mod_slowconst

#endif

#endif //AUDIO_POST_PROCESSING_H





