// Copyright(c) 2005 Analog Devices, Inc. All Rights Reserved.
// This software is proprietary and confidential to Analog Devices, Inc. and its licensors.

// File    : $Id: //depot/development/visualaudio/common/layoutsupport/2.5.0/Include/Layout.h#2 $ 
// Part of : VisualAudio V2.5.0 
// Updated : $Date: 2006/10/12 $ by $Author: Fernando $

/**
 * @file Layout.h
 * @brief Definitions for AMF_Layout.
 *
 *    Description:    Layout of post processing
 *    Copyright:        Analog Devices (c) 2003
 *    Owner:            Analog Devices
 *
 *    Revision:
 *
 *    D. Jaffe        1 May 2003        Initial Revision
 *    D. Nixon        3 May 2003        Significantly changed to use framework support functions.
 *                                    Descriptors folded to single 32 bit words. Output channel
 *                                    mapping added.
 *    D. Jaffe        7 May 2003        Renamed to Configuration.h.     Updated comments, etc.
 *    D. Jaffe        19 Jan 2004        Renamed to Layout.h.
 *    D. Jaffe        3 March 2004    Added preset fields
 *    D. Jaffe        31 March 2004    Split into LayoutManager.h/Layout.h
 */


#ifndef LAYOUT_H
#define LAYOUT_H



#include "AudioProcessing.h"
/**
 * @brief Scratch buffer description
 *
 * See description on "scratchBuffers" array in Layout structure for details.
 */
typedef struct {
    /* Size or offset */
    unsigned int size;

    /* Low 24 bits are base index (if aliased); high bits are memory spec */
    unsigned int spec;
} AMF_ScratchBufferDescript;

/**
 * AMF_Layout defines the post processing network
 *
 * It contains a table which describes the buffers that the modules reference.
 * The modules refer to these by integer index which is then resolved via
 * this table. 
 *
 * It contains the entry point for the postprocessing code, in the
 * form of an AMF_Grouping object, which in turn contains a list of modules
 * that comprise the post processing
 */
typedef struct _AMF_Layout
{
    /**
     * Description of the top-level post-processing render grouping.
     */
    AMF_Grouping topLevelGrouping;

    /** The number of static buffers that are pre-allocated by the layout.
     * These are not scratch (they retain their state).
     */
    const int staticBufferCount;

    /** The number of scratch buffers needed by the grouping.
     * Note that input and output pin counts are derived from the top level
     * grouping that is a member of this structure.
     */
    const int scratchBufferCount;

    /**
     * The following array of structs specifies size and order of
     * buffers to be allocated from a simple heap.
     *
     * If the lower 24 bits of scratchBuffers[i].spec are zero, then
     * scratchBuffers[i].size is the size of the dynamically allocated
     * buffer to be put at bufferTable[inputCount+outputCount+staticBufferCount+i].
     * The top 8 bits of scratchBuffers[i].spec indicates whether the
     * buffer should be allocated from the 'fast' (on chip dm)
     * heap, the PM heap (on chip) or the 'slow' (possibly off chip) heap.
     *
     * If the lower 24 bits of scratchBuffers[i].spec are non-zero then
     * they constitute the "baseIndex" and scratchBuffers[i].size is the
     * offset of an "aliased" buffer such that:
     *
     *   bufferTable[inputCount+outputCount+staticBufferCount+i] =
     *         bufferTable[inputCount+outputCount+staticBufferCount+(baseIndex-1)] + offset
     *
     * These arrays are in AMF_SLOW_DMDA_SEGMENT, statically allocated.
     */
    const AMF_ScratchBufferDescript *const scratchBuffers;

    /**
     * The tool statically allocates storage for all buffer pointers according
     * to the buffers used by the grouping.
     */
    AMF_Signal **const bufferTable;

    /* The following give the scratch memory usage for fast memory,  pm memory, and slow memory.
     * They are used by the framework to allocate scratch memory for the layout.
     * Note that these numbers do not include the decoder usage, nor decoder scratch
     * usage, which may be shared with layout scratch usage.  For example, if a decoder uses
     * pm scratch, the framework must allocate MAX(pmScratchCount,decoderPMScratchCount) and
     * pass this memory to AMF_Layout_Init().
     */
    int fastScratchCount, pmScratchCount, slowScratchCount;
    
    /**
     * The framework is required to assign to this pointer the address of the layout buffer
     * to be processed prior to calling AMF_Layout_Render().
     */
    AMF_Signal *pProcessingBuffer;

    /** The currently detected sample rate. */
    unsigned int iSampleRate;

    /* The next two fields are used for the layout profiler. */
    unsigned int iProfileCyclesFiltered; /* Cycles */
    unsigned int iProfileCyclesPeak;     /* Peak cycles */
    int nProfileSamplePeriods;           /* Size of profiling quantum (set by framework) */

    /* This struct is used for the profiled module. Profiling quantum is one tick */
    struct {
        unsigned int iCyclesFiltered; /* Cycles */
        unsigned int iCyclesPeak;     /* Peak cycles */
        AMF_Module *pModule;
        AMF_RenderFunction currentRenderFunc;
    } moduleProfiling;

    /* Tick counter. Incremented for each tick computed. */
    unsigned int nTicks;
} AMF_Layout;

/**
 * @brief Bits to specify the meaning of the values of the scratchBufferSpecs
 * array
 *
 * PM_HEAP indicates that the buffer should be allocated from the PM heap.
 *
 * SLOW_HEAP indicates that the buffer should be allocated from the
 * "slow" heap, which may be off chip.
 *
 * The bits masked by BASE_MASK, if non-zero, specify that the buffer is
 * an "alias" of another and give the index of the buffer number that is
 * aliased. In this case the size is actually the offset from the beginning
 * of that buffer.
 */
#define AMFScratchSpec_HEAP_SHIFT 30
#define AMFScratchSpec_PM_HEAP ((unsigned int)2<<AMFScratchSpec_HEAP_SHIFT)
#define AMFScratchSpec_SLOW_HEAP ((unsigned int)1<<AMFScratchSpec_HEAP_SHIFT)
#define AMFScratchSpec_CONTROL_HEAP ((unsigned int)3<<AMFScratchSpec_HEAP_SHIFT)
#define AMFScratchSpec_BASE_MASK 0xffffff

/**
 * @brief Initialize the AMF Config structure.
 *
 * This function initializes the AMF layout structure by
 * allocating all the scratch buffers and initializing the grouping.
 * It is normally called from FW_Init(), but may be called when
 * loading a new layout. In addition, it must be called when
 * integrating a VisualAudio layout inot a custom framework.
 *
 * @param instance            AMF layout structure to initialize.
 */
void AMF_Layout_Init(AMF_Layout *restrict instance,AMF_Signal *fastHeap,AMF_Signal pm *pmHeap,AMF_Signal *slowHeap);

/**
 * @brief Entry point for post-processing rendering for a given layout.
 *
 * This is normally an internal function, except when integrating
 * a VisualAudio layout into a custom framework.
 * Input is a set of consecutive buffer arrays in consecutive mono format.
 * Output is a set of consecutive buffer arrays as described by 'instance'.
 * Each input and output buffer has 'nSamplesPerFrame'
 * samples. On return from this call, output routing will assign the buffers in
 * 'pOutput' to the output device.
 *
 * @param instance                AMF layout structure.
 * @param nTimeSamplesPerFrame    Number of samples per frame.
 */
void AMF_Layout_Render(
        AMF_Layout *restrict instance,
        int nSamplePeriodsPerFrame);

/**
 * @brief Call this function to initialize or change nProfileSamplePeriods.
 *
 * This is normally an internal function, except when integrating
 * a VisualAudio layout into a custom framework.
 *
 * @param instance                AMF layout structure
 *
 * @return Returns a value that must be passed back to AMF_Layout_EndProfile().
 */
void AMF_Layout_InitProfile(
        AMF_Layout *restrict instance,
        int nProfileSamplePeriods);

/**
 * @brief Call this function at start of profiling interval
 *
 * This is normally an internal function, except when integrating
 * a VisualAudio layout into a custom framework.
 *
 * @param instance                AMF layout structure (unused, but included for symmetry with AMF_Layout_EndProfile)
 *
 * @return Returns a value that must be passed back to AMF_Layout_EndProfile().
 */
#define AMF_Layout_BeginProfile(_instance) AMF_GetCycles()

/**
 * @brief Call this function at end of profiling interval
 *
 * This is normally an internal function, except when integrating
 * a VisualAudio layout into a custom framework.
 *
 * @param instance                AMF layout structure.
 * @param iBeginProfileReturn    Value returned by AMF_Layout_BeginProfile
 * @param iProfilingSamplePeriods The number of sample periods of audio elapased since AMF_Layout_BeginProfile was called.
 */
void AMF_Layout_EndProfile(
        AMF_Layout *restrict instance,
        unsigned int iBeginProfileReturn);


/* If instance is non-NULL and is a member of aLayout,  
   "unprofile" any currently profiled module in aLayout,
   then profile the module by saving its currentRenderFunc.

   If instance is NULL, cancel any profiling for specified layout.

   May be called only from user level.

   @param aLayout                Layout to do profiling.
   @param instance                Module to profile or NULL.
*/
void AMF_SetModuleProfiling(AMF_Layout *aLayout,AMF_Module *instance);


/**
 * Internal utility function to compute an array of pointers to buffers
 * within a buffer group.
 * The addresses are fully defined given a pointer to the group, the number of
 * sanples in each buffer, and the buffer descriptor.
 *
 * @param pBufferArray                Pointer to the destination array.
 * @param pBuffers                    Base address of the buffer collection.
 * @param nTimeSamplesPerFrame        Number of samples in each buffer.
 * @param nBufferCount                number of buffers
 */
unsigned int AMF_BuildBufferArray(
        AMF_Signal **pBufferArray,            // pointer to destination array
        AMF_Signal *pBuffers,                // base address of buffer collection
        unsigned int nTimeSamplesPerFrame,    // number of samples in any buffer
        int nBufferCount);                    // how many buffers

/**
 * Post processing module control status values.
 */
typedef enum
{
    /** Post processing module not executed. */
    AMFModuleStatus_INACTIVE = 0,

    /** Post-processing module called normally. */
    AMFModuleStatus_ACTIVE,

    /** mute() called instead of post processing module. */
    AMFModuleStatus_MUTED,

    /** bypass() called instead of post-processing module. */
    AMFModuleStatus_BYPASSED
} AMF_ModuleStatus;

/** Module status bit mask. */
#define AMFModuleStatus_MASK        (0x00000003)

/**
 * set module status
 *
 * This function changes the module status as specified.
 *
 * @param instance          Instance to operate on
 * @param status            New status
 */
void AMF_SetModuleStatus(
    AMF_Module *instance,AMF_ModuleStatus status);



/**
 * get module status
 *
 * This function returns the module status
 *
 * @param instance          Instance to operate on
 */
AMF_ModuleStatus AMF_GetModuleStatus(AMF_Module *instance);

/**
 * sets module render function 
 *
 * This function changes the module render function to the one specified.
 * Note that any call to AMF_SetModuleStatus() overrides this value.
 *
 * @param instance          Instance to operate on
 * @param func              New function
 */
void AMF_SetModuleRenderFunction(
    AMF_Module *instance,AMF_RenderFunction func);


/**
 * gets module render function 
 *
 * Returns module render function.
 *
 * @param instance          Instance to operate on
 */
AMF_RenderFunction AMF_GetModuleRenderFunction(AMF_Module *instance);


/**
 * Render the grouping. 
 *
 * This internal function operates on the input buffer by pumping the
 * data through the rendering functions in the grouping, delivering output
 * to the output buffers.
 *
 * instance          Pointer to the grouping structure.
 * bufferList        Array of buffers to render with.
 * tickSize          Number of time samples to process. (If stereo, actual number is twice this)
 */
void AMF_Grouping_Render(
        AMF_Grouping *restrict instance,
        AMF_Signal *restrict *bufferList,
        int tickSize);


/**
 * process the grouping.
 *
 * This internal function analyzes all the modules in the group and builds the
 * renderFuncs table of pointer to functions for AMF_Grouping_Render
 * to execute.
 *
 * @param instance          Instance to process - 'this'
 */
void AMF_Grouping_Init(
        AMF_Grouping *instance);


/** Grouping class description */
extern const AMF_ModuleVariableClass AMFClassGrouping;


#endif //LAYOUT_H
