// Copyright(c) 2005 Analog Devices, Inc. All Rights Reserved.
// This software is proprietary and confidential to Analog Devices, Inc. and its licensors.

// File    : $Id: //depot/development/visualaudio/common/layoutsupport/2.5.0/Include/Tuning_Defines.h#2 $ 
// Part of : VisualAudio Designer v2.0.1
// Updated : $Date: 2006/10/12 $ by $Author: Fernando $

#ifndef _TUNING_DEFINES_H_
#define _TUNING_DEFINES_H_

#define AMF_VERSION 2 /* Version of LayoutSupport library */

/**
 * This file contains defines that are relevant to both the DSP and the host side of
 * tuning communications.  
 */

typedef enum {
    AMFTuningAction_NONE = 0,                     /* Written by DSP */
    /* Sending commands */
    AMFTuningAction_PC_SENDING_COMMAND = 3,       /* Written by PC to send a command */
    AMFTuningAction_DSP_REPLY_SENT = 4            /* DSP responds with this, where (high 16 bits give length) */
} FW_TuningAction;


/* Bits for AMFMaster.iFlags */
/**
 * This bit is set to signal the framework to not render (audio continues to run)
 * This is used in demand render mode and may also be used by the framework
 * to pause audio.  Note that the audio should produce zeros when no data
 * has been rendered.
 * This is also the value passed to AMFCommand_RENDER_CONTROL
 */
#define AMFFlag_PAUSED                     (0x00000001)

/**
 * This bit indicates that a module had the wrong version.
 */
#define AMFFlag_MODULE_VERSION_ERROR     (0x00000002)

/**
 * These bits are available to the platform to signal any platform info
 */
#define AMFFlag_PLATFORM_STATUS_MASK     (0xFFFF0000)


/* These are host commands used by the tuning interface.
 * Note that the VisualAudio framework uses these host commands for its host micro communications,
 * as well as additional commands, defined in HostProtocol_Defines.h.
 */
/** Write 32 bit storage from the message. */
#define AMFCommand_WRITE_STORAGE32        (0x0001)

/** Write 32 bit storage from the message atomically. */
#define AMFCommand_WRITE_STORAGE32_ATOMIC    (0x0002)

/** Read 32 bit storage. */
#define AMFCommand_READ_STORAGE32        (0x0003)

/** Read 32 bit storage. */
#define AMFCommand_READ_STORAGE32_ATOMIC (0x0004)

/*
 * Writes structured data
 */
#define AMFCommand_WRITE_STRUCTURED            (0x0005)
/* Obsolete name... */
#define AMFCommand_WRITE_STRUCTURED32        (0x0005)

/** Controls rendering (values are 0 = audio driver-driven, 1 = host demand render driven.
 *  VisualAudio framework overrides this.
 */
#define AMFCommand_RENDER_CONTROL             (0x0006)

/*
 * Arg is layout index.
 * Returns: buffer pointer
 */
#define AMFCommand_QUERY_PROCESSING_BUFFER (0x0007)

/** Execute AMF_Layout_Render() from foreground.
 * Arguments are layout index and number of samples to render
 */
#define AMFCommand_RENDER                    (0x0008)

/*
 * Arg is layout index.
 * Returns: sample rate (integer)
 */
#define AMFCommand_QUERY_SAMPLE_RATE        (0x0009)

/* 16-bit commands are only supported for Blackfin */
/** Write 16 bit storage from the message. */
#define AMFCommand_WRITE_STORAGE16        (0x000a)

/** Write 16 bit storage from the message atomically. */
#define AMFCommand_WRITE_STORAGE16_ATOMIC    (0x000b)

/** Run interrupt profiler */
#define AMFCommand_RUN_INTERRUPT_PROFILING (0x000c)

/** Call AMF_SetModuleProfiling(arg1,arg2) */
#define AMFCommand_SET_MODULE_PROFILING (0x000d) 

#define AMFCommand_APPLY_PRESET (0x000e)

/** Error return codes */
/** Success return code. */
#define E_NO_ERROR                    (0x0000)

/** These are errors detected on the DSP side */
/** The DSP-received message CRC was wrong. */
#define E_CRC_ERROR                    (0x0001)

/** The command opcode is not implemented. */
#define E_NOT_IMPLEMENTED            (0x0002)

/** The message was too long. CRC and opcode were not checked. */
#define E_TOO_LONG                    (0x0003)

/** The message count is wrong and exceeds the message length. */
#define E_COUNT_ERROR                (0x0004)

/** The operation requested could not be performed. */
#define E_FAILED                    (0x0005)

/** Invalid layout */
#define E_INVALID_LAYOUT            (0x0006)

/** Invalid preset (un-supported op code found) */
#define E_INVALID_PRESET            (0x0007)

/** PC-only error codes begin here */
#define E_PC_BASE                   (0x0100)



/* Maybe move from here down to a separate file? */
#define MAX_MSG_SIZE 0x3fff
/* Can't ever be bigger than this, because of the number of bits we've allocated
 * However, the actual maximum allowable size depends on what was allocated by the
 * tool in the macro AMF_MAX_MSG_SIZE, which is written to ToolGeneratedLayouts.h
 */


/** Return value from a tuning or host message handler that does not handle the message. */
#define FW_NOT_HANDLED                            0

/** Return value from a tuning or host message handler that does handle the message. */
#define FW_HANDLED                                1

/**
 * Return value from a tuning or host command message handler that has built a reply
 * and wants it sent.
 */
#define FW_REPLY_BUILT                            0x20000

/**
 * Return value from a tuning or host command message handler that wants to reply
 * to a command message with 'code'.
 */
#define FW_REPLY_WITH(code)                    (0x10000 | code)

/**
 * Used to tag host messages in the message queue that originated from the host.
 * Such messages need a reply.
 * Any tagged message not handled by the Framework, and also not handled
 * by the user control code results in a reply of E_NOT_IMPLEMENTED.
 */
#define FW_REPLY_REQUIRED                        0x80000000

/**
 * Build a message header.
 *
 * This macro constructs a dword for the header of a command or reply
 * packet.
 *
 * length            Length of the packet.
 * code                Command/reply code.
 * return            Header dword.
 */
#define FW_MakeMessageHeader(length, code)        (((length) << 16) | (code))

/**
 * Extract the length from a message header dword.
 *
 * This macro extracts the packet length located in the high 16 bits of the
 * header dword.
 *
 * hdr                Header dword.
 * return            Length of the packet.
 */
#define FW_GetMessageLength(hdr)                (((hdr) >> 16) & MAX_MSG_SIZE)


/**
 * Extract the command/reply code from the header dword.
 *
 * This macro extracts the packet command/reply code located in the low 16 bits
 * of the header dword.
 *
 * hdr                Header dword.
 * return            Command/reply code.
 */
#define FW_GetMessageType(hdr)                    ((hdr) & 0xFFFF)

/**
 * Structured setting "op-codes", specified in the upper bits of the address
 * These are used for both presets and AMFCommand_WRITE_STRUCTURED32
 */
#define AMFStructuredSetting_ARRAY 0
#define AMFStructuredSetting_SPARSE_ARRAY 1
#define AMFStructuredSetting_STATUS 2
#define AMFStructuredSetting_ARRAY16 3 /* Blackfin only */

/* Macro for constructing a preset or structured setting */
#define AMFStructuredSetting_OP_SHIFT 16
#define AMF_StructuredSetting_OP_SHIFT 16 // Old mis-named version

#define AMF_MakeStructuredSettingAddress(_sym,_offset) ((unsigned int)((((char *)&(_sym)))+((unsigned int)(_offset))))

#define AMFStructuredSetting_ATOMIC ((unsigned)(1<<31)) /* If this bit is set in setting descriptor, disable interrupts */


/* The following must be kept in synch with AudioProcessing.h.
 * They are used by the host to find the module-specific render variables
 */
#ifndef AMF_Module_SIZE
#define AMF_Module_SIZE         5
#define AMF_ModuleVariable_SIZE     (AMF_Module_SIZE+4)
#endif

#endif
