// Copyright(c) 2005 Analog Devices, Inc. All Rights Reserved.
// This software is proprietary and confidential to Analog Devices, Inc. and its licensors.

// File    : $Id: //depot/development/visualaudio/common/layoutsupport/2.5.0/Source/AMF_Grouping.c#2 $ 
// Part of : VisualAudio V2.5.0 
// Updated : $Date: 2006/10/12 $ by $Author: Fernando $

/**
 * @addtogroup Framework    Framework Core
 * @{
 */

/**
 * @file AMF_Grouping.c
 * @brief Post processing module
 *
 *    Description:    Post-processing initialization and module walking.
 *    Copyright:        Analog Devices (c) 2003
 *    Owner:            Analog Devices
 *
 *    Revision:
 *
 *    D. Jaffe        1 May 2003        Initial Revision
 *    D. Jaffe        5 May 2003        Support for bypass, mute, etc., changed some names.
 *  M. Nixon       27 May 2003        Added doxygen comments.
 *  D. Jaffe       3  Aug 2003        Changed class structure definition.
 *                                  Changed MAX_BUFFERS from 16 to 32.
 *  D. Jaffe       11 Sep 2003        Changed MAX_BUFFERS to use alloca()
 *  D. Jaffe       25 Sep 2003        Fixed bug in status initialization
 */

#include <string.h>
#include "Tuning.h"


#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
SEG_MOD_SLOW_CODE void AMF_Grouping_Init(
        AMF_Grouping *grouping)
{
    /* Set up render functions */
    int i, j, k;
    int numModules = grouping->numModules;
    const unsigned int *initialStatusBitVectors = grouping->initialStatusBitVectors;
    if (!initialStatusBitVectors)
    {
        /* Simple case: Make everything ACTIVE */
        for (i = 0; i < numModules; i++)
        {
            AMF_SetModuleStatus(grouping->modules[i], AMFModuleStatus_ACTIVE);
        }
    }
    else
    {
        /* Use info in array */
        for (k = 0, j = 0, i = 0; i < numModules; i++)
        {
            AMF_SetModuleStatus(grouping->modules[i],
                                (AMF_ModuleStatus)((initialStatusBitVectors[k] >> j)
                                                   & AMFModuleStatus_MASK));
            j += 2; // Increment must equal log2(AMFModuleStatus_MASK+1)
            if (j == 32)
            {
                k++;
                j = 0;
            }
        }
    }
#ifdef AMF_SIMPLE_BUFFERING
    {
        /* Layout.c has already initialized the input and output pins in bufferTable at this point */
        AMF_Module *module;
          AMF_Signal **layoutBufferTable = grouping->layout->bufferTable;
        int bufferIndexCount;
        const int *bufferIndecies;
        for (i = 0; i < numModules; i++) {
            module = grouping->modules[i];
            bufferIndexCount = module->bufferIndexCount;
            bufferIndecies = module->bufferIndecies;
            for (j = 0; j < bufferIndexCount; j++)
                ((AMF_Signal **)(bufferIndecies))[j] = layoutBufferTable[bufferIndecies[j]];
        }
    }
#endif
}

/**
 * @brief Perform custom bypass for the grouping module.
 *
 * If a grouping module is placed in bypass mode, then its function
 * pointer will be set to point to this custom bypass function. The bypass
 * functions job is to copy the data from the input pins to the output pins.
 *
 * @param instance            Pointer to the instance structure.
 * @param buffers            Buffers to process.
 * @param tickSize            Number of time samples to process.
 */
#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
/*ARGSUSED*/
static SEG_MOD_SLOW_CODE void AMF_Grouping_Bypass(
        AMF_Grouping *restrict grouping,
        AMF_Signal *restrict *bufferList, 
        int tickSize)
{   
    if (!grouping->b.b.grouping) {
        /*
         * We're the top level grouping.
         * This is special-purpose code
         * that takes advantage of the knowledge
         * that the post-processing is done in place
         * and that the buffers are contiguous.
         * Bypassing the top level grouping is as
         * simple as zeroing any outputs that don't
         * have corresponding inputs.  
         */
        int numInputs, numOutputs;
        numInputs = grouping->b.inputPinCount;
        numOutputs = grouping->b.outputPinCount;
        if (numOutputs > numInputs)
            memset(bufferList[numInputs * 2], 0,
                   tickSize * (numOutputs-numInputs) * sizeof(AMF_Signal));
    } else {
        /* It's not the top level grouping */
        /* Hiearchical groupings are not implemented yet */
    }
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_speed
#endif
/*ARGSUSED*/
SEG_MOD_FAST_CODE void AMF_Grouping_Render(
        AMF_Grouping *restrict grouping,
        AMF_Signal *restrict *bufferList, 
        int tickSize)
{
    int i;
    AMF_Module *module;
    AMF_Module * const *modules = grouping->modules;
    int numModules = grouping->numModules;
    int myTickSize = grouping->tickSize;
    AMF_RenderFunction rf;

#ifndef AMF_SIMPLE_BUFFERING
    int bufferIndexCount,j;
    const int *bufferIndecies;
    // Array for buffers passed to module
    // We use VisualDSP++'s dynamic array allocation off the stack.
    // Alternative: Could allocate from our heap
    AMF_Signal *moduleBufferlist[grouping->maxBufferCount];    
    // In bypass and mute modes, the inputs and outputs are used by generic
    // functions. If we ever do "compiled groups", then the inputs and outputs
    // will be propogated to children here.    Otherwise, inputs and outputs are ignored.
    AMF_Signal * *layoutBufferTable = grouping->layout->bufferTable;
#endif

    // Eventually, when/if we support hierarchical grouping with different tick
    // sizes, we will combine the passed-in tickSize with myTickSize to determine
    // how many times to call children.
    // Execute render functions of children
    for (i = 0; i < numModules; i++)
    {
        module = modules[i];
        rf = module->currentRenderFunc;
#ifdef AMF_SIMPLE_BUFFERING
        (rf)(module, (AMF_Signal **)module->bufferIndecies, myTickSize);
#else
        bufferIndexCount = module->bufferIndexCount;
        bufferIndecies = module->bufferIndecies;
        for (j = 0; j < bufferIndexCount; j++)
        {
            moduleBufferlist[j] = layoutBufferTable[bufferIndecies[j]];
        }
        (rf)(module, moduleBufferlist, myTickSize);
#endif
    }
    {
        AMF_FreqDomainInfo *fdi = grouping->freqDomainInfo;
        if (fdi) {
            if (fdi->ticksToNextHop == 0)
                fdi->ticksToNextHop = fdi->ticksPerHop;
            fdi->ticksToNextHop--; /* Counts down from ticksPerHop-1 to 0 */
        }
    }
}

SEG_MOD_FAST_CONST const AMF_ModuleVariableClass AMFClassGrouping = 
{
    AMFModuleClassFlag_VARIABLE_PIN_COUNT,   // flags
    (AMF_RenderFunction)AMF_Grouping_Render, // default render function
    (AMF_RenderFunction)AMF_Grouping_Bypass  // custom bypass
};

/**
 * @}
 *
 * End of file.
 */

