// Copyright(c) 2005 Analog Devices, Inc. All Rights Reserved.
// This software is proprietary and confidential to Analog Devices, Inc. and its licensors.

// File    : $Id: //depot/development/visualaudio/common/layoutsupport/2.5.0/Source/Layout.c#2 $ 
// Part of : VisualAudio V2.5.0 
// Updated : $Date: 2006/10/12 $ by $Author: Fernando $

/**
 * @addtogroup Framework    Framework Core
 * @{
 */

/**
 * @file Layout.c
 * @brief Layout manager module
 *
 *    Description:    Post-processing buffer management and render dispatch.
 *    Copyright:        Analog Devices (c) 2003
 *    Owner:            Analog Devices
 *
 *    Revision:
 *
 *    D. Jaffe        1 May 2003        Initial Revision
 *    D. Nixon        3 May 2003        Significantly changed to use framework support functions.
 *    D. Jaffe        1 June 2003        Changed AMF_Config_Init() to support base/offset style sub-allocations.
 *                                  Also changed it to optimize_for_space
 *    D. Jaffe        2 June 2003        Changed to call AMF_Grouping_Render() indirectly.
 *                                  That way post-processing can be set to mute or bypass as needed.
 *    D. Jaffe        5 Aug 2003        Changed the allocation of the post-processing buffer.
 *    D. Jaffe        17 Sep 2003        Changed to init heaps (moved heaps to config)
 *    D. Jaffe        19 Jan 2004        Changed name to Layout.c
 *    E. Yin            22 Jan 2005        Changed the algorithm to avoid using of run time lib
 */


#include <string.h>
#include "Tuning.h"
#include "Layout.h"

#ifndef MAX
#define MAX(x,y) ((x)>(y)?(x):(y))
#endif
#ifndef MIN
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif




#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
SEG_MOD_FAST_CODE unsigned int AMF_BuildBufferArray(
        AMF_Signal ** pBufferArray,                // pointer to destination array
        AMF_Signal * pBuffers,                    // base address of buffer collection
        unsigned int nTimeSamplesPerFrame,    // number of samples in any buffer
        int nBufferCount)                    // how many buffers
{
    int i;
    unsigned int bufferIndex = 0;
    unsigned int offs = 0;
    for (i = 0; i < nBufferCount; i++)
    {
        // Set the buffer pointer for the Nth buffer.
        pBufferArray[bufferIndex++] = &pBuffers[offs];

        // Move the offset to the next buffer.
        offs += nTimeSamplesPerFrame;
    }

    /* Return the number of buffer pointers written. */
    return nBufferCount;
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
SEG_MOD_FAST_CODE void AMF_CalculateProfile(
    unsigned int *iProfileCyclesFiltered,
    unsigned int *iProfileCyclesPeak,
    unsigned int iCyclesStart,
    unsigned int iCyclesEnd) {
    /* This is an internal function used to calculate profile */
    /* Wrapping is not an issue since 2s complement takes care of it for us */
    float fCycles;
    unsigned int iCycles;
    unsigned int savedMode1;
    /* We want the following subtraction to be non-saturating (i.e. wrapping) */
#ifdef __ADSPBLACKFIN__
    iCycles = iCyclesEnd - iCyclesStart; /* On BF, this generates a non-saturating subtraction */
#else
    /* On SHARC, must set/restore mode around subtraction, to get non-saturation */
    asm volatile("%0 = mode1; bit clr mode1 8192; nop; %1 = %2 - %3; mode1 = %0;":"+d" (savedMode1), "=&d" (iCycles) : "d" (iCyclesEnd), "d" (iCyclesStart));
#endif
    fCycles = (float)iCycles;
#define PROFILE_FILTER_COEFF 0.9
    *iProfileCyclesFiltered = (int)((float)*iProfileCyclesFiltered * PROFILE_FILTER_COEFF +
                                    fCycles * (1.0-PROFILE_FILTER_COEFF));
    *iProfileCyclesPeak = MAX(iCycles,*iProfileCyclesPeak);
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_speed
#endif
static SEG_MOD_FAST_CODE void profile(AMF_Module *restrict instance,
        AMF_Signal *restrict *bufferList, int tickSize)
{
    unsigned int startCycles;
    AMF_Grouping *grouping = instance->grouping;
    AMF_Layout *layout = grouping->layout;
    startCycles = AMF_GetCycles();    
    layout->moduleProfiling.currentRenderFunc(instance,bufferList,tickSize);
    AMF_CalculateProfile(&layout->moduleProfiling.iCyclesFiltered,
                         &layout->moduleProfiling.iCyclesPeak,
                         startCycles,AMF_GetCycles());
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_speed
#endif
static SEG_MOD_FAST_CODE void getPinTypes(bool isIndirect, unsigned int *bitVectorPtr,
                                          int numPins, int *pinTypes, int *typeBitMask) 
{ 
    int i,j;
    unsigned int bitVect;
    *typeBitMask = 0;
    if (numPins == 0) /* Needed or loop below fails */
        return;
    if (isIndirect) {
        int bitVectIndex;
        unsigned int nibble;
        for (bitVectIndex = 0, i=0; i<numPins; bitVectIndex++) {
            bitVect = bitVectorPtr[bitVectIndex];
            for (j = 0; (j<8) && (i<numPins); i++, j++, bitVect >>= 4) {
                nibble = bitVect & 0xf;
                pinTypes[i] = nibble;
                *typeBitMask |= (1<<nibble);
            }
         }
    } else {
        int pinType,lim;
        lim = MIN(numPins,8);
        bitVect = bitVectorPtr[0];
        for (i=0; i<lim; i++, bitVect >>= 4) {
            pinTypes[i] = bitVect & 0xf;
            *typeBitMask |= (1<<pinTypes[i]);
        }
        /* Remainder use the last value */
#pragma no_vectorization // We don't know alignment of the stack array
        for (pinType = pinTypes[lim-1]; i<numPins; i++) 
            pinTypes[i] = pinType;
    }
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_speed
#endif
static SEG_MOD_FAST_CODE void bypassAux(AMF_Signal *restrict *bufferList, 
                                        int numInputs,int numOutputs,
                                        int *inputPinTypes, 
                                        int *outputPinTypes,
                                        int desiredType,
                                        int size) {
    AMF_Signal *src, *dst;
    int lim,i,inputNum, outputNum;
    bool foundIt;
    int inputPinType,outputPinType;
    outputNum = 0;
    for (inputNum=0; inputNum<numInputs && outputNum<numOutputs; inputNum++) {
        inputPinType = inputPinTypes[inputNum];
        if (inputPinType == desiredType) {
            /* We got an input pin of desired type */
            for (foundIt = false; (!foundIt) && (outputNum<numOutputs); outputNum++) {
                outputPinType = outputPinTypes[outputNum];
                if (outputPinType == desiredType) {
                    /* We got an output pin of desired type */
                    foundIt = true;
                    src = bufferList[inputNum];
                    dst = bufferList[outputNum+numInputs];
                    if (src != dst)
                        memcpy(dst, src, size*sizeof(AMF_Signal));
                }
            }
        }
    }
    /* Mute remainder */
    for (; outputNum<numOutputs; outputNum++) {
        outputPinType = outputPinTypes[outputNum];
        if (outputPinType == desiredType) {
            dst = bufferList[numInputs+outputNum];
            memset(dst, 0, size*sizeof(AMF_Signal));
        }
    }
}

/**
 * @brief Perform bypass for a post-processing module.
 *
 * If a post-processing module is placed in bypass mode, then its function
 * pointer will be set to point to this generic bypass function. The bypass
 * functions job is to copy the data from the input pins to the output pins.
 *
 * @param instance            Pointer to the instance structure.
 * @param buffers            Buffers to process.
 * @param tickSize            Number of time samples to process.
 */
#ifndef NO_OPTIMIZE
#pragma optimize_for_speed
#endif
static SEG_MOD_FAST_CODE void bypassRender(AMF_Module *restrict instance,
        AMF_Signal *restrict *bufferList, int tickSize)
{
    /* we'll have a two pass algorithm.
       During the first pass, we'll deal with audio pins.
       During the second pass, we'll deal with control pins.
       The kth input audio (control) pin will be mapped to the kth output audio (control pin).
       If a mono audio input maps to a stereo output output, we'll duplicate the channels.
       If a stereo audio input maps to a mono stereo output, we'll calculate (L+R)/2
    */
    const AMF_ModuleClass *bc = instance->classInfo;
    int numInputs, numOutputs;
    unsigned int *inputPinBitVectorPtr, *outputPinBitVectorPtr;
    bool inputIsIndirect = ((bc->flags & AMFModuleClassFlag_INDIRECT_INPUT_PIN_TYPE) != 0);
    bool outputIsIndirect = ((bc->flags & AMFModuleClassFlag_INDIRECT_OUTPUT_PIN_TYPE) != 0);
    if (bc->flags & AMFModuleClassFlag_VARIABLE_PIN_COUNT)
    {
        AMF_ModuleVariable *biv = (AMF_ModuleVariable *)instance;
        numInputs = biv->inputPinCount;
        numOutputs = biv->outputPinCount;
        if (inputIsIndirect)
            inputPinBitVectorPtr = ((unsigned int *)biv->inputPinBitVectorOrPointer);
        else 
            inputPinBitVectorPtr = (unsigned int *)&biv->inputPinBitVectorOrPointer;
        if (outputIsIndirect)
            outputPinBitVectorPtr = ((unsigned int *)biv->outputPinBitVectorOrPointer);
        else 
            outputPinBitVectorPtr = (unsigned int *)&biv->outputPinBitVectorOrPointer;
    }
    else
    {
        numInputs = bc->inputPinCount;
        numOutputs = bc->outputPinCount;
        if (inputIsIndirect)
            inputPinBitVectorPtr = ((unsigned int *)bc->inputPinBitVectorOrPointer);
        else 
            inputPinBitVectorPtr = (unsigned int *)&bc->inputPinBitVectorOrPointer;
        if (outputIsIndirect)
            outputPinBitVectorPtr = ((unsigned int *)bc->outputPinBitVectorOrPointer);
        else 
            outputPinBitVectorPtr = (unsigned int *)&bc->outputPinBitVectorOrPointer;
    }
    {
        int inputPinTypes[numInputs];    
        int outputPinTypes[numOutputs];    
        AMF_Signal *src, *dst;
        int lim,i,inputNum, outputNum;
        bool foundIt;
        int inputPinType,outputPinType;
        int dummy,outputTypeBitMask; /* to see if any bits of each type are present */
        int fftSize = (instance->grouping) ? instance->grouping->freqDomainInfo->fftSize : tickSize;
        int sizeForPinType[AMFPinType_PIN_TYPE_COUNT] = 
             {tickSize, tickSize*2, 1, fftSize, fftSize*2, fftSize/2+2, (fftSize/2+1)*2};

        getPinTypes(inputIsIndirect, inputPinBitVectorPtr, numInputs, inputPinTypes, &dummy);
        getPinTypes(outputIsIndirect, outputPinBitVectorPtr, numOutputs, outputPinTypes, &outputTypeBitMask);

        for (i=AMFPinType_CONTROL; i<AMFPinType_PIN_TYPE_COUNT; i++)
            /* Note that we don't do PCM (MONO and STEREO) here */
            if (outputTypeBitMask & (1<<i))
                bypassAux(bufferList, numInputs, numOutputs, inputPinTypes,
                          outputPinTypes, i, sizeForPinType[i]);

        if (!(outputTypeBitMask & ((1<<AMFPinType_MONO) | (1<<AMFPinType_STEREO))))
            return; /* No PCM output pins */

        /* We have output PCM */
        /* Now repeat for pcm signals, which are handled specially because of mono/stereo conversion */
        for (outputNum=0,inputNum=0; inputNum<numInputs && outputNum<numOutputs; inputNum++) {
            inputPinType = inputPinTypes[inputNum];
            if (inputPinType == AMFPinType_MONO || inputPinType == AMFPinType_STEREO) {
                /* We got an input pcm pin */
                for (foundIt = false; (!foundIt) && (outputNum<numOutputs); outputNum++) {
                    outputPinType = outputPinTypes[outputNum];
                    if (outputPinType == AMFPinType_MONO || outputPinType == AMFPinType_STEREO) {
                        foundIt = true;
                        src = bufferList[inputNum];
                        dst = bufferList[numInputs+outputNum];
                        if (outputPinType == inputPinType) {
                            if (src != dst)
                                memcpy(dst, src, sizeForPinType[outputPinType]*sizeof(AMF_Signal));
                        } else if (outputPinType == AMFPinType_STEREO) {
                            /* mono to stereo.  Duplicate */
                            for (i=0; i<tickSize; i++) {
                                dst[i*2] = src[i];
                                dst[i*2+1] = src[i];
                            }
                        } else /* stereo to mono. Mix */ {
                            lim = sizeForPinType[AMFPinType_STEREO];
                            for (i=0; i<lim; i+=2) {
#ifdef __ADSPBLACKFIN__
                                *dst++ = ((src[i]>>1) + (src[i+1]>>1)); 
#else                            
                                *dst++ = (src[i] + src[i+1]) * 0.5;
#endif
                            }
                        }
                    } 
                }
            }
        }
        /* Mute remaining pcm pins */
        for (; outputNum<numOutputs; outputNum++) {
            outputPinType = outputPinTypes[outputNum];
            if (outputPinType == AMFPinType_MONO || outputPinType == AMFPinType_STEREO) {
                dst = bufferList[numInputs+outputNum];
                memset(dst, 0, sizeForPinType[outputPinTypes[outputNum]] * sizeof(AMF_Signal));
            }
        }
    }
}

/**
 * @brief Perform mute for a post-processing module.
 *
 * If a post-processing module is placed in mute mode, then its function
 * pointer will be set to point to this generic bypass function. The mute
 * functions job is to write silence to the output pins.
 *
 * @param instance            Pointer to the instance structure.
 * @param buffers            Buffers to process.
 * @param tickSize            Number of time samples to process.
 */
#ifndef NO_OPTIMIZE
#pragma optimize_for_speed
#endif
static SEG_MOD_FAST_CODE void muteRender(AMF_Module *restrict instance,
        AMF_Signal *restrict *bufferList, int tickSize)
{
    // Clear outputs of any module
    const AMF_ModuleClass *bc = instance->classInfo;
    bool outputIsIndirect = ((bc->flags & AMFModuleClassFlag_INDIRECT_OUTPUT_PIN_TYPE) != 0);
    int numInputs, numOutputs;
    unsigned int *outputPinBitVectorPtr;
    if (bc->flags & AMFModuleClassFlag_VARIABLE_PIN_COUNT)
    {
        AMF_ModuleVariable *biv = (AMF_ModuleVariable *)instance;
        numInputs = biv->inputPinCount;
        numOutputs = biv->outputPinCount;
        if (outputIsIndirect)
            outputPinBitVectorPtr = ((unsigned int *)biv->outputPinBitVectorOrPointer);
        else 
            outputPinBitVectorPtr = (unsigned int *)&biv->outputPinBitVectorOrPointer;
    }
    else
    {
        numInputs = bc->inputPinCount;
        numOutputs = bc->outputPinCount;
        if (outputIsIndirect)
            outputPinBitVectorPtr = ((unsigned int *)bc->outputPinBitVectorOrPointer);
        else 
            outputPinBitVectorPtr = (unsigned int *)&bc->outputPinBitVectorOrPointer;
    }
    {
        int outputPinTypes[numOutputs];
        int i,dummy;
        int fftSize = (instance->grouping) ? instance->grouping->freqDomainInfo->fftSize : tickSize;
        int sizeForPinType[AMFPinType_PIN_TYPE_COUNT] = 
             {tickSize, tickSize*2, 1, fftSize, fftSize*2, fftSize/2+2, (fftSize/2+1)*2};
        getPinTypes(outputIsIndirect, outputPinBitVectorPtr, numOutputs, outputPinTypes, &dummy);
        for (i = 0; i < numOutputs; i++)
            memset(bufferList[i + numInputs], 0, sizeForPinType[outputPinTypes[i]] * sizeof(AMF_Signal));
    }
}

/**
 * @brief Perform inactive for a post-processing module.
 *
 * Null function.
 */
#ifndef NO_OPTIMIZE
#pragma optimize_for_speed
#endif
SEG_MOD_FAST_CODE void AMF_InactiveRender(AMF_Module *restrict instance,
        AMF_Signal *restrict *bufferList, int tickSize)
{
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
static SEG_MOD_SLOW_CODE AMF_RenderFunction *getModuleRenderFunctionPtr(AMF_Module *instance) {
    AMF_Grouping *grouping = instance->grouping;
    AMF_Layout *layout;
    /*
     * It's possible that someone will use the modules in stand-alone fashion and there
     * will be no grouping structure.
     */
    if (!grouping)
        return &instance->currentRenderFunc;
    else
        layout = grouping->layout;
    return ((layout->moduleProfiling.pModule == instance) ?
            &layout->moduleProfiling.currentRenderFunc :
            &instance->currentRenderFunc);
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
SEG_MOD_SLOW_CODE AMF_RenderFunction AMF_GetModuleRenderFunction(AMF_Module *instance) {
    AMF_RenderFunction *crf = getModuleRenderFunctionPtr(instance);
    return *crf;
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
SEG_MOD_SLOW_CODE AMF_ModuleStatus AMF_GetModuleStatus(
    AMF_Module *instance) {
    AMF_RenderFunction crf = AMF_GetModuleRenderFunction(instance);
    return (crf == AMF_InactiveRender ? AMFModuleStatus_INACTIVE :
            crf == muteRender ? AMFModuleStatus_MUTED :
            crf == ((instance->classInfo->bypassFunction) ?
                     (instance->classInfo->bypassFunction) :
                     bypassRender) ? AMFModuleStatus_BYPASSED :
            AMFModuleStatus_ACTIVE);
    /* Note that the render function might not be instance->classInfo->renderFunction,
       as that is only the default render function */
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
SEG_MOD_SLOW_CODE void AMF_SetModuleRenderFunction(
        AMF_Module *instance, 
        AMF_RenderFunction newRenderFunc) {
    AMF_RenderFunction *crf = getModuleRenderFunctionPtr(instance);
    *crf = newRenderFunc;
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
SEG_MOD_SLOW_CODE void AMF_SetModuleStatus(
        AMF_Module *instance, 
        AMF_ModuleStatus status)
{
    AMF_RenderFunction *crf = getModuleRenderFunctionPtr(instance);
    int layoutSupportVersion = (instance->classInfo->flags &
                                AMFModuleClassFlag_LAYOUT_SUPPORT_VERSION_MASK) >> 16;
    if (layoutSupportVersion > AMF_VERSION) {
        AMFMaster.iFlags |= AMFFlag_MODULE_VERSION_ERROR;
        status = AMFModuleStatus_INACTIVE;
    }
    switch (status)
    {
    case AMFModuleStatus_ACTIVE:
        *crf = instance->classInfo->renderFunction;
        break;
    case AMFModuleStatus_MUTED:
        *crf = muteRender;
        break;
    case AMFModuleStatus_BYPASSED: 
        *crf = ((instance->classInfo->bypassFunction) ?
                (instance->classInfo->bypassFunction) :
                bypassRender);
        break;
    case AMFModuleStatus_INACTIVE:
    default:
        *crf = AMF_InactiveRender;
        break;
    }
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
SEG_MOD_SLOW_CODE void AMF_SetModuleProfiling(AMF_Layout *aLayout,AMF_Module *instance) {
    /* If instance is non-NULL and is a member of aLayout,  
       "unprofile" any currently profiled module in aLayout,
       then profile the module by saving its currentRenderFunc.

       If instance is NULL, cancel any profiling for specified layout.

       May be called only from user level.
    */
    if ((!aLayout) || (instance && (instance->grouping->layout != aLayout)))
        return; /* Invalid or wrong layout */
    /* Disable interrupts, in case AMF_SetModuleRenderFunction()
       or other function called at interrupt level
     */
     AMF_BeginAtomic();
    if (aLayout->moduleProfiling.pModule)
        aLayout->moduleProfiling.pModule->currentRenderFunc = aLayout->moduleProfiling.currentRenderFunc;
    if (instance) {
        aLayout->moduleProfiling.pModule = instance;
        aLayout->moduleProfiling.currentRenderFunc = instance->currentRenderFunc;
        instance->currentRenderFunc = profile;
        aLayout->moduleProfiling.iCyclesFiltered = 0;
        aLayout->moduleProfiling.iCyclesPeak = 0;
    } else {
        aLayout->moduleProfiling.pModule = NULL;
        aLayout->moduleProfiling.currentRenderFunc = NULL;
    }
    AMF_EndAtomic();
}


/*
 * This function perform post-processing by calling AMF_Grouping_Render with the input data, which
 * is written to the output data. On return from this function, the output data is written to the
 * mapped output channels by the framework.
 *
 * The size of the input and output buffers is always an integer multiple of AMF_TICK_SIZE. When
 * that multiple is not 1, AMF_Grouping_Render is called as many times as the mulltiple, with the
 * pointers advanced by AMF_TICK_SIZE between calls.
 */
#ifndef NO_OPTIMIZE
#pragma optimize_for_speed
#endif
SEG_MOD_FAST_CODE void AMF_Layout_Render(
        AMF_Layout *restrict instance,    // the object controlling the rendering - 'this'
        int nTimeSamples)                // number of time samples in supplied buffer.
                                        // It is guaranteed to be an integer multiple of the tick size
{
#ifndef AMF_SIMPLE_BUFFERING
    int samplesProcessed;
    AMF_Signal *pSampsStart;
#endif    
    AMF_Signal **pBuffTable = &instance->bufferTable[0];
    AMF_RenderFunction rf;
    int tickSize = instance->topLevelGrouping.tickSize;
    rf = instance->topLevelGrouping.b.b.currentRenderFunc;
    if (!rf) /* Inactive? */ 
        return;
#ifdef AMF_SIMPLE_BUFFERING
    (rf)(&instance->topLevelGrouping.b.b, pBuffTable, tickSize);
    instance->nTicks++;
#else    
    // Loop processing all samples in the frames.
    for (samplesProcessed = 0; samplesProcessed < nTimeSamples; samplesProcessed += tickSize)
    {
        pSampsStart = &instance->pProcessingBuffer[samplesProcessed];

        /* Build the array of input buffer pointers first. */
        AMF_BuildBufferArray(&pBuffTable[0],
                            pSampsStart,
                            nTimeSamples,
                            instance->topLevelGrouping.b.inputPinCount);

        /* Build the array of output buffer pointers next. */
        AMF_BuildBufferArray(&pBuffTable[instance->topLevelGrouping.b.inputPinCount],
                            pSampsStart,
                            nTimeSamples,
                            instance->topLevelGrouping.b.outputPinCount);

        /* Process the samples. */
        (rf)(&instance->topLevelGrouping.b.b, pBuffTable, tickSize);
        instance->nTicks++;
    }
#endif
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_speed
#endif
SEG_MOD_SLOW_CODE void AMF_Layout_InitProfile(AMF_Layout *restrict instance,int nProfileSamplePeriods) {
    instance->nProfileSamplePeriods = nProfileSamplePeriods;
    instance->iProfileCyclesFiltered = 0;
    instance->iProfileCyclesPeak = 0;
}

// AMF_Layout_BeginProfile() is a macro, defined in Layout.h:  
// #define AMF_Layout_BeginProfile(_instance) AMF_GetCycles()

#ifndef NO_OPTIMIZE
#pragma optimize_for_speed
#endif
SEG_MOD_SLOW_CODE void AMF_Layout_EndProfile(AMF_Layout *restrict instance,unsigned int startCycles) {
    AMF_CalculateProfile(&instance->iProfileCyclesFiltered,&instance->iProfileCyclesPeak,
                         startCycles,AMF_GetCycles());
}

/* forward decl */
int AMF_ProfileInterrupts(unsigned int startCycles,unsigned int threshold);

/*
 * This function allocates scratch storage for the grouping, and then initializes it.
 */
#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
SEG_MOD_SLOW_CODE void AMF_Layout_Init(AMF_Layout *restrict instance,AMF_Signal *fastScratch,AMF_Signal pm *pmScratch,AMF_Signal *slowScratch)
{
    /* instance->scratchBufferSpecification is interpreted as follows:
     * 
     * If the top 8 bits are zero, scratchBufferSpecification[i] indicates
     * the size of the malloc'ed buffer to be put at bufferTable[inputCount+outputCount+i].
     * If the top 8 bits are non-zero, scratchBuffersSpecification[i] & 0xffffff 
     * and (scratchBufferSpecification[i] >> 24) are the offset and baseIndex respectively
     * such that:
     *
     *   bufferTable[inputCount+outputCount+i] = bufferTable[inputCount+outputCount+(baseIndex-1)] +
     *                                           offset
     */ 
    int i;
    unsigned int size,base,offset,spec,heapInfo,j;
    int scratchStart = (instance->topLevelGrouping.b.inputPinCount
                        + instance->topLevelGrouping.b.outputPinCount
                        + instance->staticBufferCount);
    AMF_Signal **pBuffTable = &instance->bufferTable[scratchStart];
    // The locations [0:(scratchStart-1)] are set dynamically before each tick
    int scratchBufferCount = instance->scratchBufferCount;

    /*
     * Make two passes.  On the first pass, fill in all the true allocations.
     * On the second pass, fill in the aliases.  By making two passes,
     * we can allow forward references. 
     */
    for (i = 0; i < scratchBufferCount; i++)
    {
        spec = instance->scratchBuffers[i].spec; 
        base = spec & AMFScratchSpec_BASE_MASK;
        if (!base) { /* Allocate it */
            size = instance->scratchBuffers[i].size;
            heapInfo = spec >> AMFScratchSpec_HEAP_SHIFT;
            switch (heapInfo) {
                case (AMFScratchSpec_SLOW_HEAP>>AMFScratchSpec_HEAP_SHIFT):
                    pBuffTable[i] = slowScratch;
                    slowScratch += size;
                    break;
                case (AMFScratchSpec_PM_HEAP>>AMFScratchSpec_HEAP_SHIFT):
                    pBuffTable[i] = pmScratch;
                    pmScratch += size;
                    break;
                case (AMFScratchSpec_CONTROL_HEAP>>AMFScratchSpec_HEAP_SHIFT):
                    /* This is always the last one and there's always
                      * only one, so we can do this safely
                      * without worrying about subsequent scratch buffers
                       * overwriting our pBuffTable[].
                      */
                    for (j=0; j<size; j++) 
                        pBuffTable[i+j] = fastScratch++;
                    break;
                default:
                    pBuffTable[i] = fastScratch;
                    fastScratch += size;
                    break;
            }
        }
    }
    for (i = 0; i < scratchBufferCount; i++)
    {
        spec = instance->scratchBuffers[i].spec; 
        base = spec & AMFScratchSpec_BASE_MASK;
        if (base) {
//          ASSERT(pBuffTable[base-1]) 
            offset = instance->scratchBuffers[i].size; /* It's really the offset into an already-allocated block in this case */
            pBuffTable[i] = pBuffTable[base-1] + offset;
        }
    }
#ifdef AMF_SIMPLE_BUFFERING
    /* Build the array of input buffer pointers first. */
    AMF_BuildBufferArray(&instance->bufferTable[0],
                         instance->pProcessingBuffer,
                         instance->topLevelGrouping.tickSize,
                         instance->topLevelGrouping.b.inputPinCount);

    /* Build the array of output buffer pointers next. */
    AMF_BuildBufferArray(&instance->bufferTable[instance->topLevelGrouping.b.inputPinCount],
                         instance->pProcessingBuffer,
                         instance->topLevelGrouping.tickSize,
                         instance->topLevelGrouping.b.outputPinCount);
#endif    
    AMF_Grouping_Init(&instance->topLevelGrouping);

    /* Finally, run interrupt profiler with interrupts disabled to determine loop length.
     * Would be better to do this somewhere else that's global to all layouts, but there's
     * currently no AMF_Init().
     */

// any number that's big enough to average out set-up/tear-down */
#define PROFILE_INTERRUPTS_CALIBRATION_CYCLES 400000 
    {
        int x;
        // Profile with interrupts disabled
        AMF_BeginAtomic();
        x = AMF_ProfileInterrupts(AMF_GetCycles(),PROFILE_INTERRUPTS_CALIBRATION_CYCLES);
        AMF_EndAtomic();
#ifdef __ADSPBLACKFIN__
        // divide calibration_cycles by x, rounding the result
        AMFMaster.iInterruptProfileLoopLength = (PROFILE_INTERRUPTS_CALIBRATION_CYCLES + (x/2))/x;
#else
        // divide calibration_cycles by x, rounding the result
        AMFMaster.iInterruptProfileLoopLength = (int)(PROFILE_INTERRUPTS_CALIBRATION_CYCLES/(float)x + 0.5);
#endif //__ADSPBLACKFIN__
    }
    instance->nTicks = 0;
}

/**
 * @}
 *
 * End of file.
 */

