// Copyright(c) 2005 Analog Devices, Inc. All Rights Reserved.
// This software is proprietary and confidential to Analog Devices, Inc. and its licensors.

// File    : $Id: //depot/development/visualaudio/common/layoutsupport/2.5.0/Source/Tuning.c#2 $ 
// Part of : VisualAudio V2.5.0 
// Updated : $Date: 2006/10/12 $ by $Author: Fernando $

asm("#include \"processor.h\"");
#include "Tuning.h"
#include <string.h>

/**
 * @brief CRC32 algorithm.
 *
 * This is the classic CRC algorithm that iterates once per message bit. When
 * applied to a message with a CRC included in the length, it returns 0 making
 * validation easy. 
 *
 * @param data                   Pointer to the packet.
 * @param len                    Length of the packet (including CRC word).
 */
/* We optimize for speed because this function is called at interrupt level */
#ifdef __ADSPBLACKFIN__
/* calcCRC32() for blackfin is in Utilities.asm */
unsigned int calcCRC32(unsigned int *data,unsigned int length);
#else

// SHARC version
#ifndef NO_OPTIMIZE
#pragma optimize_for_speed
#endif
static SEG_MOD_FAST_CODE unsigned int calcCRC32(
        unsigned int *data,
        unsigned int length)
{

    unsigned int i;
    unsigned int sum;
    unsigned int savedMode1;

    asm volatile("%0 = mode1;":"=d" (savedMode1));
    asm volatile("bit clr mode1 8192;"); // turn off saturation

    sum = 0;
#if defined(__ADSP21367__) || defined(__ADSP21368__) || defined(__ADSP21369__)
    /* Can't use SIMD here, because tuning buffer might be off-chip */
#pragma no_vectorization
#else    
    /* Assume it's OK to use SIMD. */
#pragma vector_for
#endif    
    for (i = 0; i < length; i++)
    {
        sum += data[i];
    }
    
    // WARNING: for some reason, the optimizer puts this inside
    // various checks, so it looks like it might not happen with 
    // length == 1
    asm volatile("mode1 = %0;"::"d" (savedMode1));

    return (~sum + 1);
}
#endif

/**
 * @brief Comm driver Message CRC verification function
 *
 * Returns TRUE if the CRC of the message in pCommandBuff is valid.
 *
 * @return                TRUE if CRC was valid
 */
#ifndef NO_OPTIMIZE
#pragma optimize_for_speed
#endif
SEG_MOD_FAST_CODE bool AMF_VerifyCRC(unsigned int *pCommandBuff)
{
    int length = FW_GetMessageLength(pCommandBuff[0]);
    if (length < 2)
    {
        /* Received messages less than 2 words long are always an error. */
        return false;
    }
    return (calcCRC32(pCommandBuff, length) == 0);
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
static SEG_MOD_SLOW_CODE AMF_Layout *getLayout(int index) {
    if (index < AMFMaster.nLayoutCount)
        return AMFMaster.pLayouts[index];
    return NULL;
}

/* This applies a preset or structured setting, as addressed by nAddress.
 * nAddress is passed to pGetFunc, which increments it and returns the value at that
 * address. pContext is passed uninterpreted to getFunc
 *
 * Returns error code.
 */
#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
SEG_MOD_SLOW_CODE int AMF_ApplyPreset(int nAddress,AMF_GetPresetElementFunc getFunc,void *pContext, bool bMaskInterrupts){
    int nDescriptionCount; // Number of descriptions 
    int n;                 // Description or value to apply
    int nBlockAddr;        // Base address for applying of the value
    int count;             // Number of elements in an array or sparse array
    int i,j;               // Loop counters
    int err = E_NO_ERROR;  // Return value
    int bDisableInterrupts = 0;
    int bInterruptsDisabled = 0;

#define GET_NEXT_VALUE(_address,_value) if ((err=getFunc(&(_address),&(_value),pContext))!=E_NO_ERROR) goto goHome

    //get the description count and increment the address pointer
    GET_NEXT_VALUE(nAddress, nDescriptionCount); //assume the description count is always the first data
    if (bMaskInterrupts) {
        AMF_BeginAtomic();
    }
    if (nDescriptionCount > 0xFFFF) { // Assume bogus 
        err = E_INVALID_PRESET;
        goto goHome;
    }

    for (i = 0; i < nDescriptionCount; i++)
    {
        //read opcode word first
        GET_NEXT_VALUE(nAddress, n);
        //next read target address
        GET_NEXT_VALUE(nAddress, nBlockAddr);
        
        if (!bMaskInterrupts) {
            bDisableInterrupts = n & AMFStructuredSetting_ATOMIC;
            if (bDisableInterrupts != bInterruptsDisabled) {
                if (!bInterruptsDisabled) {
                    AMF_BeginAtomic();
                  } else {
                    AMF_EndAtomic();
                  }
                  bInterruptsDisabled = bDisableInterrupts;
            }
        }
        //Assume nBlockAddr is valid (no easy way to validiate it)
        //check the op code
        switch((n >> 16) & 0x7fff) // Cf. Tuning_Defines.h
        {
            default: { /* If this happens, could be a garbage preset */
                err = E_INVALID_PRESET;
                goto goHome;
            }
            case AMFStructuredSetting_STATUS:
                // low half of op code is a status value
                //apply the status value
                AMF_SetModuleStatus((AMF_Module *)(((int*)(nBlockAddr))),(AMF_ModuleStatus)(n&0xffff));
                break;
            case AMFStructuredSetting_ARRAY:
                // low half of op is a count
                count = n & 0xffff;
                for (j = 0; j < count; j++)
                {    
                    //read the next 32-bit word
                    GET_NEXT_VALUE(nAddress, n);
                    //apply the setting by writting the preset value into DSP internal memory
                    ((int*)(nBlockAddr))[j] = n;
                }
                break;
#ifdef __ADSPBLACKFIN__
            case AMFStructuredSetting_ARRAY16: {
                // low half of op is a count
                short s;
                count = n & 0xffff;
                j = 0; 
                while (j < count)
                {    
                    //read the next pair of 16 bit words
                    GET_NEXT_VALUE(nAddress, n);
                    //apply the setting by writting the preset value into DSP internal memory
                    s = (short)(n >> 16);
                    ((short*)(nBlockAddr))[j] = s;
                    j++;
                    if (j < count) { 
                        /* If j == count, then there are an odd number of items and
                         * this is the last one.
                         */
                        s = (short)(n & 0xffff);
                        ((short*)(nBlockAddr))[j] = s;
                    }
                    j++;
                }
                break;
            }
#endif
         case AMFStructuredSetting_SPARSE_ARRAY:
            {    
                //Complicated case...
                // low half of op is a count, followed by a skip bit vector
                int j, skipbit;
                int offset;  //address offset for DSP internal memory
                count = n & 0xffff;
                //read skip bit vector
                GET_NEXT_VALUE(nAddress, skipbit);
                //always apply to the first address
                GET_NEXT_VALUE(nAddress, n);
                (*(int*)(nBlockAddr)) = n;
                count--; //descrement the count by 1
                offset = 1;  //since we always set the first address
                for (j = 0; (j < count); offset++)
                {    
                    //checks the skip bit vector to see if we need to
                    //set the data for the address.
                    //After 32 checks, assume all data should be set.
                    if (offset > 32 || (!((skipbit >> (offset - 1)) & 0x1)))
                    {
                        //read the settings
                        GET_NEXT_VALUE(nAddress, n);
                        ((int*)(nBlockAddr))[offset] = n;
                        j++;  //updated the un-skipped address
                    }
                }
                break;
            }
        }
    }
 goHome:
    if ( bMaskInterrupts || bInterruptsDisabled) 
    {
        AMF_EndAtomic();
    }
    return err;
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
static SEG_MOD_SLOW_CODE int getNextCommandBuffElement(int *nAddress,int *pResult,void *pContext) {
    /* Helper function passed to AMF_ApplyPreset */
    int nEndAddress = ((int)pContext);
    int nCurAddr = *nAddress;
    if (nCurAddr > nEndAddress) /* nEndAddress is the last valid address */
        return E_COUNT_ERROR;
    *pResult = *((int *)nCurAddr);
    *nAddress = nCurAddr + sizeof(int);   /* Increment for next time */
    return E_NO_ERROR;
}

#pragma optimize_for_speed
SEG_MOD_FAST_CODE int AMF_ProfileInterrupts(unsigned int startCycles,unsigned int threshold) {
    int i;
    unsigned int cycles;
    i = 0;
    do {
//        inlining doesn't always happen, so we need to do it by hand
//        cycles = AMF_GetCycles();
#ifndef __ADSPBLACKFIN__
          __asm volatile ("%0=EMUCLK;" : "=d" (cycles));
#else
          __asm volatile ("%0=CYCLES;" : "=d" (cycles));
#endif
        i = i + 1;
    }
    while (cycles - startCycles < threshold); /* This works ok even if clock wraps once */
    return i;
}

#define setReplyLengthAndStatus(pCommandBuff,length, status) \
    pCommandBuff[0] = FW_MakeMessageHeader(length,status)


/*
 * This function handles tuning command messages (a subset of host messages),
 *
 * The return convention is:
 *        FW_NOT_HANDLED (= 0)        the framework does not handle this message
 *        FW_REPLY_WITH(value):        return a defualt 16-bit reply value to the host
 *        FW_REPLY_BUILT:                the reply is in AMFMaster.pCommandBuff, reply with that
 */
#ifndef NO_OPTIMIZE    
#pragma optimize_for_space
#endif
SEG_MOD_SLOW_CODE int AMF_CommHandler(
            unsigned int *pCommandBuff,
            unsigned int msgLength,
            unsigned int nCommandBuffSize
            )
{
    int type = FW_GetMessageType(pCommandBuff[0]);
    /* We could verify length here if desired. - DAJ */
        
    /* Only examine host command messages. */
    switch (type) {

       case AMFCommand_WRITE_STRUCTURED:            
        {
            /* These commands write structured data, as defined elsewhere. */
            int err;
            /*
             * Invoke the structued write (same format as a preset).
             * We pass the end address of the valid data (not including CRC) as the
             * optional "context" argument. 
             */
            err = AMF_ApplyPreset((int)&pCommandBuff[1],
                                  (AMF_GetPresetElementFunc)getNextCommandBuffElement,
                                  (void *)&pCommandBuff[msgLength-2], false); 
            return FW_REPLY_WITH(err);
        }

        case AMFCommand_WRITE_STORAGE32_ATOMIC:
        case AMFCommand_WRITE_STORAGE32:
            /* These commands write up to AMFMaster.pCommandBuff_SIZE - 3 words. The layout is:
               address, arg1, ..., argN.  (The -3 is to account for the header word,
               address, and CRC.)
            */
            {
                unsigned int *pDest = (unsigned int *)pCommandBuff[1];
                 bool isAtomic = false;
                 if (type == AMFCommand_WRITE_STORAGE32_ATOMIC)
                     isAtomic = true;
//                bool isAtomic = (type == AMFCommand_WRITE_STORAGE32_ATOMIC); // This caused crash in BF compiler (fixed?)
                if (isAtomic)
                {
                    DISABLE_INTERRUPTS(); /* Atomic writes require interrupts disabled */
                }
                memcpy(pDest, &pCommandBuff[2], (msgLength-3)*sizeof(int));
                if (isAtomic)
                {
                    ENABLE_INTERRUPTS();
                }
                return FW_REPLY_WITH(E_NO_ERROR);
            }

        case AMFCommand_READ_STORAGE32:
        case AMFCommand_READ_STORAGE32_ATOMIC:
        /* This command reads up to pCommandBuff_SIZE - 2 words. The reply layout is:
           arg1, ..., argN (The -2 is to account for the header word and CRC).
        */
            if (pCommandBuff[2] <= (unsigned int)(nCommandBuffSize - 2))
            {
                unsigned int *pSrc = (unsigned int *)pCommandBuff[1];
                unsigned int nLength = pCommandBuff[2];
// Crashes BF compiler:                bool isAtomic = (type == AMFCommand_READ_STORAGE32_ATOMIC);
                 bool isAtomic = false;
                 if (type == AMFCommand_READ_STORAGE32_ATOMIC)
                     isAtomic = true;
                if (isAtomic)
                {
                    DISABLE_INTERRUPTS(); /* Atomic writes require interrupts disabled */
                }
                memcpy(&pCommandBuff[1], pSrc, nLength*sizeof(int));
                if (isAtomic)
                {
                    ENABLE_INTERRUPTS(); /* Atomic writes require interrupts disabled */
                }
                setReplyLengthAndStatus(pCommandBuff, nLength + 2, E_NO_ERROR);
                return FW_REPLY_BUILT;
            }
            return FW_REPLY_WITH(E_COUNT_ERROR);

#ifdef __ADSPBLACKFIN__
        case AMFCommand_WRITE_STORAGE16_ATOMIC:
        case AMFCommand_WRITE_STORAGE16:
            /* These commands write up to AMFMaster.pCommandBuff_SIZE - 4 words. The layout is:
               address, count, arg1, ..., argN.  (The -4 is to account for the header word,
               address, count and CRC.)
            */
            {
                unsigned int *pDest = (unsigned int *)pCommandBuff[1];
                 bool isAtomic = false;
                int count = pCommandBuff[2];
                if (count > (int)(msgLength)-4)
                    return FW_REPLY_WITH(E_COUNT_ERROR);
                 if (type == AMFCommand_WRITE_STORAGE16_ATOMIC)
                     isAtomic = true;
//                bool isAtomic = (type == AMFCommand_WRITE_STORAGE16_ATOMIC); // This causes crash in BF compiler
                if (isAtomic)
                {
                    DISABLE_INTERRUPTS(); /* Atomic writes require interrupts disabled */
                }
                memcpy(pDest, &pCommandBuff[3], count*sizeof(short));
                if (isAtomic)
                {
                    ENABLE_INTERRUPTS();
                }
                return FW_REPLY_WITH(E_NO_ERROR);
            }
#endif

        case AMFCommand_QUERY_SAMPLE_RATE: 
        case AMFCommand_QUERY_PROCESSING_BUFFER: {
        /* This command returns the sample rate. */
            AMF_Layout *pLayout = getLayout(pCommandBuff[1]);
            if (!pLayout)
                return FW_REPLY_WITH(E_INVALID_LAYOUT);
            pCommandBuff[1] = ((type == AMFCommand_QUERY_SAMPLE_RATE) ? 
                                 pLayout->iSampleRate : 
                                 (unsigned long)pLayout->pProcessingBuffer);
            setReplyLengthAndStatus(pCommandBuff, 3, E_NO_ERROR);
            return FW_REPLY_BUILT;
        }                

        case AMFCommand_RENDER_CONTROL:  // Only set/clear pause bit here. Framework does the rest
            if (pCommandBuff[1])
                AMFMaster.iFlags |= AMFFlag_PAUSED;
            else
                AMFMaster.iFlags &= ~AMFFlag_PAUSED;
            return FW_REPLY_WITH(E_NO_ERROR);

        case AMFCommand_RENDER: {
        /* Execute AMF_Layout_Render() once.
         * The framework must be stopped or paused for this to work properly.
           */
            AMF_Layout *pLayout = getLayout(pCommandBuff[1]);
            if (!pLayout)
                return FW_REPLY_WITH(E_INVALID_LAYOUT);
            AMF_Layout_Render(pLayout,pCommandBuff[2]);
            return FW_REPLY_WITH(E_NO_ERROR);
        }

        case AMFCommand_RUN_INTERRUPT_PROFILING: {
            /* User passes in the number of cycles it wishes for
               us to count. We return the number of iteartions
               of our loop that we were able to do before that
               cycle total was exceeded. It is necessary for the
               caller to know the size of the loop to make sense of this.
            */
            pCommandBuff[1] = (AMF_ProfileInterrupts(AMF_GetCycles(),(unsigned int)pCommandBuff[1]) *
                               AMFMaster.iInterruptProfileLoopLength); 
            setReplyLengthAndStatus(pCommandBuff, 3, E_NO_ERROR);
            return FW_REPLY_BUILT;
        }

        case AMFCommand_SET_MODULE_PROFILING: 
            AMF_SetModuleProfiling((AMF_Layout *)pCommandBuff[1],(AMF_Module *)pCommandBuff[2]);
            return FW_REPLY_WITH(E_NO_ERROR);

        case AMFCommand_APPLY_PRESET: { 
            /* The specified preset (an integer offset) to apply */
            int err;
            err = SYS_ApplyPreset(pCommandBuff[1]);
            setReplyLengthAndStatus(pCommandBuff, 2, err);
            return FW_REPLY_BUILT;
        }
    }

    /* Otherwise, this was not a tuning message. (Should never happen) */
    return FW_NOT_HANDLED;
}

/**
 * This calculates the message CRC, returns length
 */
//If we don't disable interrupts, we can optimize for space here 
#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
SEG_MOD_SLOW_CODE int AMF_ConstructReply(unsigned int *pCommandBuff,int length)
{
    unsigned int crc;

    // Calculate the CRC32 of everything up to the CRC.
    pCommandBuff[length - 1] = 0;
    crc = calcCRC32(pCommandBuff, length);

    // Write the CRC value.
    pCommandBuff[length - 1] = crc;
    return length;
}

#ifndef NO_OPTIMIZE
#pragma optimize_for_space
#endif
int AMF_HandleTuningCommand(void) {
    /* Called when we have a tuning message in tuning mode */
    int reply;
    int length;

    // Store toggle bit
    AMF_SaveToggleBitValue(AMFMaster.pCommandBuff[0]);

    if ((int)(FW_GetMessageLength(AMFMaster.pCommandBuff[0])) > AMFMaster.nCommandBuffSize)
    {
        /* It was too long. */
        setReplyLengthAndStatus(AMFMaster.pCommandBuff, 2, E_TOO_LONG);
    }
    else if (!AMF_VerifyCRC(AMFMaster.pCommandBuff))
    {
        /* the message has an error. */
        setReplyLengthAndStatus(AMFMaster.pCommandBuff, 2, E_CRC_ERROR);
    }
    else
    {
        /* Otherwise, the command is valid
         */
        reply = AMF_CommHandler(AMFMaster.pCommandBuff,
                                FW_GetMessageLength(AMFMaster.pCommandBuff[0]),
                                AMFMaster.nCommandBuffSize);
        
        /* This message requires a reply. We reply according to how the message handlers
           dealt with it. */
        if (reply == FW_NOT_HANDLED)
        {
            /* Invalid message */
            setReplyLengthAndStatus(AMFMaster.pCommandBuff, 2, E_NOT_IMPLEMENTED);
        }
        else if ((reply & FW_REPLY_BUILT) == 0)
        {
            /* This was FW_REPLY_WITH(code). Do that. */
            setReplyLengthAndStatus(AMFMaster.pCommandBuff, 2, reply & 0xffff);
        }
    }

    //Set the toggle bit before transmitting
    length = FW_GetMessageLength(AMFMaster.pCommandBuff[0]);
    AMF_RestoreToggleBitValue(AMFMaster.pCommandBuff[0]);
    return (AMF_ConstructReply(AMFMaster.pCommandBuff,length) << 16) | AMFTuningAction_DSP_REPLY_SENT;
}

