function [OUT,u] = Nonlinear_Compensator_Simplified(w,Le_tab,Bl_tab,C_tab, dLe_tab,L2_0,R2_0,M,Re,Rm,fs,table_size,max_x,stepsize_table);
    %Linear Feedback compensator (w,a,b,c,Le,Bl,C,L2_0,R2_0,M,Re,Rt,fs);
    %   u - compensated output voltage
    %   w - input voltage
    %   Bl_tab - force factor vector, tabel in mm, +-10mm
    %   Le_tab - inductance vector, tabel in mm, +-10mm
    %   C_tab - compliense vector, tabel in mm, +-10mm
    %   a - inductance polynomes, vector
    %   c - compliance polynomes, vector
    %   b - force factor polynomes, vector
    %   M - total mass
    %   Re - Electrical resistance
    %   Rm - Mechanical resistance
    %   R2 - electrical resistance due the eddy current losses
    %   fs - sampling rate
     
     Ts = 1/fs;                      % Stepsize
    Le_0 = Le_tab(table_size/2);    % inductance, converting to Henry
    Bl_0 = Bl_tab(table_size/2);    % force factor 
    C_0 = C_tab(table_size/2);      % compliance, converting to m/N
    K_0 = 1/C_0;                    % stiffness N/m
        
    fase = 0;
    v = zeros(1,length(w));  %   LD(linear dynamics) output vector
    u = zeros(1,length(w));  %   compensator output vector 
    X = zeros(3,length(w));  %   X-vector      
    
    for k=1:fase+1
    
        X(:,k)= [  0           %   x1 - current   
                   0           %   x3 - excution
                   0  ];       %   x4 - velosity  
    end

    for k=1:64000
        %Loading nonlinear parameters from tables
        index_realvalue = X(2,k)/stepsize_table + table_size/2;
        index = round(index_realvalue-0.5);
        dx  = index_realvalue - index;
         
        Bl_diff  = Bl_tab(index+1) - (Bl_tab(index));
        Le_diff  = Le_tab(index+1) - (Le_tab(index));
        C_diff   = C_tab(index+1) - (C_tab(index));
        K_diff   = 1/C_tab(index+1) - 1/C_tab(index); 
        dLe_diff = dLe_tab(index+1) - (dLe_tab(index));
      
        Bl  = Bl_tab(index)  + dx*Bl_diff;
        Le  = Le_tab(index)  + dx*Le_diff;
        C   = C_tab(index)   + dx*C_diff;
        K   = 1/C_tab(index) + dx*K_diff;
        dLe = dLe_tab(index) + dx*dLe_diff;
        
        dBl  = Bl_diff/stepsize_table;
        ddLe = dLe_diff/stepsize_table;
        dK   = K_diff/stepsize_table;
       
        %constants for speaker model
        L2 = Le*L2_0/Le_0;       
        R2 = Le*R2_0/Le_0;
        dL2 = dLe*L2_0/Le_0; 
        
        %compensator
        u(k) = Le/(Bl+dLe*X(1,k))*(X(3,k)*(X(2,k)*dK + (K-K_0) - dBl*X(1,k) ...
              - ddLe*X(1,k)^2/2 - Bl_0^2/Le_0) + Re*X(2,k)*(K-K_0)/Le_0 - ...
              Re*X(1,k)*(2*Bl+dLe*X(1,k))/(2*Le_0) + Bl_0*w(k)/Le_0) + Re*X(1,k) + ...
              Bl*X(3,k) + dLe*X(1,k)*X(3,k);% + R2*(X(1,k)-X(2,k));   
%                    u(k) =w(k);             
        
        %calculation of F and G matrix values
        F=[(1-((Re/Le)*Ts)) 0 -Ts/Le*(X(1,k)*dLe+Bl); 
            0 1 Ts;  
            (Ts*Bl/M) (-Ts/(C*M)) (1-Rm*Ts/M)];
            G=[Ts/Le ; 0 ; 0];
        %calculating the state variables
        X(:,k+1) = F*X(:,k) + G*u(k);
    
        
    end
    OUT(1,:) = X(1,:);
    OUT(3,:) = X(2,:);
    OUT(4,:) = X(3,:);
end